/*
	Basic Analog PAD plugin for PCSX Gamecube
	by emu_kidid based on the DC/MacOSX HID plugin

	TODO: Rumble?
*/

#include <gccore.h>
#include <stdint.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <ogc/pad.h>
#include "plugins.h"
#include "PsxCommon.h"
#include "PSEmu_Plugin_Defs.h"

#ifdef HW_RVL
#include <wiiuse/wpad.h>
#endif

extern void SysPrintf(char *fmt, ...);

/* Button Bits */
#define PSX_BUTTON_TRIANGLE ~(1 << 12)
#define PSX_BUTTON_SQUARE 	~(1 << 15)
#define PSX_BUTTON_CROSS	~(1 << 14)
#define PSX_BUTTON_CIRCLE	~(1 << 13)
#define PSX_BUTTON_L2		~(1 << 8)
#define PSX_BUTTON_R2		~(1 << 9)
#define PSX_BUTTON_L1		~(1 << 10)
#define PSX_BUTTON_R1		~(1 << 11)
#define PSX_BUTTON_SELECT	~(1 << 0)
#define PSX_BUTTON_START	~(1 << 3)
#define PSX_BUTTON_DUP		~(1 << 4)
#define PSX_BUTTON_DRIGHT	~(1 << 5)
#define PSX_BUTTON_DDOWN	~(1 << 6)
#define PSX_BUTTON_DLEFT	~(1 << 7)


#ifdef HW_RVL
/****************************************************************************
 * WPAD_Stick  (taken from Snes9XGX)
 *
 * Get X/Y value from Wii Joystick (classic, nunchuk) input
 ***************************************************************************/
s8 WPAD_Stick(u8 chan, u8 right, int axis)
{
	float mag = 0.0;
	float ang = 0.0;
	WPADData *data = WPAD_Data(chan);

	switch (data->exp.type)
	{
		case WPAD_EXP_NUNCHUK:
		case WPAD_EXP_GUITARHERO3:
			if (right == 0)
			{
				mag = data->exp.nunchuk.js.mag;
				ang = data->exp.nunchuk.js.ang;
			}
			break;

		case WPAD_EXP_CLASSIC:
			if (right == 0)
			{
				mag = data->exp.classic.ljs.mag;
				ang = data->exp.classic.ljs.ang;
			}
			else
			{
				mag = data->exp.classic.rjs.mag;
				ang = data->exp.classic.rjs.ang;
			}
			break;

		default:
			break;
	}

	/* calculate x/y value (angle need to be converted into radian) */
	if (mag > 1.0) mag = 1.0;
	else if (mag < -1.0) mag = -1.0;
	double val;

	if(axis == 0) // x-axis
		val = mag * sin((M_PI * ang)/180.0f);
	else // y-axis
		val = mag * cos((M_PI * ang)/180.0f);

	return (s8)(val * 128.0f);
}

#endif

/* Controller type, later do this by a Variable in the GUI */
int controllerType = 0; // 0 = standard, 1 = analog (analog fails on old games)

long  PadFlags = 0;

long PAD__init(long flags) {
	SysPrintf("start PAD_init()\r\n");

//	printf("Flags: %08x\n", flags);
	PadFlags |= flags;

	/* Read Configuration here */

	SysPrintf("end PAD_init()\r\n");

	return PSE_PAD_ERR_SUCCESS;
}

long PAD__shutdown(void) {
	return PSE_PAD_ERR_SUCCESS;
}

long PAD__open(void)
{
	return PSE_PAD_ERR_SUCCESS;
}

long PAD__close(void) {
	return PSE_PAD_ERR_SUCCESS;
}

long PAD__readPort1(PadDataS* pad) {

	int b = PAD_ButtonsHeld(0);
	uint16_t pad_status = 0xFFFF;	//bit pointless why is this done this way?

	/* PAD Buttons, Start = Start, Select = Start+Z, Cross = A, Square = B, Triangle = Y, Circle = X */
	if ((b & PAD_BUTTON_START) && (!(b & PAD_TRIGGER_Z)))
		pad_status &= PSX_BUTTON_START;
	if ((b & PAD_BUTTON_START) && (b & PAD_TRIGGER_Z))
		pad_status &= PSX_BUTTON_SELECT;
	if (b & PAD_BUTTON_A)
		pad_status &= PSX_BUTTON_CROSS;
	if (b & PAD_BUTTON_B)
		pad_status &= PSX_BUTTON_SQUARE;
	if (b & PAD_BUTTON_X)
		pad_status &= PSX_BUTTON_CIRCLE;
	if (b & PAD_BUTTON_Y)
		pad_status &= PSX_BUTTON_TRIANGLE;
	if (b & PAD_BUTTON_UP)
		pad_status &= PSX_BUTTON_DUP;
	else if (b & PAD_BUTTON_DOWN)
		pad_status &= PSX_BUTTON_DDOWN;
	if (b & PAD_BUTTON_LEFT)
		pad_status &= PSX_BUTTON_DLEFT;
	else if (b & PAD_BUTTON_RIGHT)
		pad_status &= PSX_BUTTON_DRIGHT;
	/* Shoulder Buttons, L1 = L, R1 = R, L2 = L+Z, R2 = R+Z */
	if ((b & PAD_TRIGGER_Z) && (b & PAD_TRIGGER_R))
		pad_status &= PSX_BUTTON_R2;
	if ((b & PAD_TRIGGER_Z) && (b & PAD_TRIGGER_L))
		pad_status &= PSX_BUTTON_L2;
	if ((!(b & PAD_TRIGGER_Z)) && (b & PAD_TRIGGER_R))
		pad_status &= PSX_BUTTON_R1;
	if ((!(b & PAD_TRIGGER_Z)) && (b & PAD_TRIGGER_L))
		pad_status &= PSX_BUTTON_L1;

#ifdef HW_RVL
  int h = WPAD_ButtonsHeld(0);
	/* PAD Buttons */
	if (h & WPAD_CLASSIC_BUTTON_PLUS)
		pad_status &= PSX_BUTTON_START;
	if (h & WPAD_CLASSIC_BUTTON_MINUS)
		pad_status &= PSX_BUTTON_SELECT;
	if (h & WPAD_CLASSIC_BUTTON_B)
		pad_status &= PSX_BUTTON_CROSS;
	if (h & WPAD_CLASSIC_BUTTON_Y)
		pad_status &= PSX_BUTTON_SQUARE;
	if (h & WPAD_CLASSIC_BUTTON_A)
		pad_status &= PSX_BUTTON_CIRCLE;
	if (h & WPAD_CLASSIC_BUTTON_X)
		pad_status &= PSX_BUTTON_TRIANGLE;
	if (h & WPAD_CLASSIC_BUTTON_UP)
		pad_status &= PSX_BUTTON_DUP;
	else if (h & WPAD_CLASSIC_BUTTON_DOWN)
		pad_status &= PSX_BUTTON_DDOWN;
	if (h & WPAD_CLASSIC_BUTTON_LEFT)
		pad_status &= PSX_BUTTON_DLEFT;
	else if (h & WPAD_CLASSIC_BUTTON_RIGHT)
		pad_status &= PSX_BUTTON_DRIGHT;
	/* Shoulder Buttons */
	if (h & WPAD_CLASSIC_BUTTON_FULL_R)
		pad_status &= PSX_BUTTON_R2;
	if (h & WPAD_CLASSIC_BUTTON_FULL_L)
		pad_status &= PSX_BUTTON_L2;
	if (h & WPAD_CLASSIC_BUTTON_ZR)
		pad_status &= PSX_BUTTON_R1;
	if (h & WPAD_CLASSIC_BUTTON_ZL)
		pad_status &= PSX_BUTTON_L1;
#endif

  if(!controllerType) {
    //allow GC analog pad to be used for digital
    if(PAD_StickX(0) < -50)
      pad_status &= PSX_BUTTON_DLEFT;
    else if(PAD_StickX(0) > 50)
      pad_status &= PSX_BUTTON_DRIGHT;
    if(PAD_StickY(0) < -50)
      pad_status &= PSX_BUTTON_DDOWN;
    else if(PAD_StickY(0) > 50)
      pad_status &= PSX_BUTTON_DUP;
#ifdef HW_RVL
    //allow Classic Controller analog pad to be used for digital
    if(WPAD_Stick(0,0,0) < -50)
      pad_status &= PSX_BUTTON_DLEFT;
    else if(WPAD_Stick(0,0,0) > 50)
      pad_status &= PSX_BUTTON_DRIGHT;
    if(WPAD_Stick(0,0,1) < -50)
      pad_status &= PSX_BUTTON_DDOWN;
    else if(WPAD_Stick(0,0,1) > 50)
      pad_status &= PSX_BUTTON_DUP;
#endif
  	pad->controllerType = PSE_PAD_TYPE_STANDARD; 	// Standard Pad
  }

  else if(controllerType==1) {
  	//adjust values by 128 cause psx values in range 0-255 where 128 is center position
#ifndef HW_RVL
  	pad->leftJoyX  = (u8)(PAD_StickX(0)+127) & 0xFF;				//analog stick
  	pad->leftJoyY  = (u8)(-PAD_StickY(0)+127) & 0xFF;
  	pad->rightJoyX = (u8)(PAD_SubStickX(0)+127) & 0xFF;			//C-stick (Left JoyStick)
  	pad->rightJoyY = (u8)(PAD_SubStickY(0)+127) & 0xFF;
#else
  	pad->leftJoyX  = ((u8)(PAD_StickX(0)+127) + (u8)(WPAD_Stick(0,0,0)+127)) & 0xFF;
  	pad->leftJoyY  = ((u8)(-PAD_StickY(0)+127) + (u8)(-WPAD_Stick(0,0,1)+127)) & 0xFF;
  	pad->rightJoyX = ((u8)(PAD_SubStickX(0)+127) + (u8)(WPAD_Stick(0,1,0)+127)) & 0xFF;
  	pad->rightJoyY = ((u8)(PAD_SubStickY(0)+127) + (u8)(WPAD_Stick(0,1,1)+127)) & 0xFF;
#endif
    pad->controllerType = PSE_PAD_TYPE_ANALOGPAD; 	// Analog Pad  (Right JoyStick)
  }

	pad->buttonStatus = pad_status;					//Copy Buttons
	return PSE_PAD_ERR_SUCCESS;
}

long PAD__readPort2(PadDataS* pad) {

	int b = PAD_ButtonsHeld(1);

	uint16_t pad_status = 0xFFFF;	//bit pointless why is this done this way?

	/* PAD Buttons, Start = Start, Select = Start+Z, Cross = A, Square = B, Triangle = Y, Circle = X */
	if ((b & PAD_BUTTON_START) && (!(b & PAD_TRIGGER_Z)))
		pad_status &= PSX_BUTTON_START;
	if ((b & PAD_BUTTON_START) && (b & PAD_TRIGGER_Z))
		pad_status &= PSX_BUTTON_SELECT;
	if (b & PAD_BUTTON_A)
		pad_status &= PSX_BUTTON_CROSS;
	if (b & PAD_BUTTON_B)
		pad_status &= PSX_BUTTON_SQUARE;
	if (b & PAD_BUTTON_X)
		pad_status &= PSX_BUTTON_CIRCLE;
	if (b & PAD_BUTTON_Y)
		pad_status &= PSX_BUTTON_TRIANGLE;
	if (b & PAD_BUTTON_UP)
		pad_status &= PSX_BUTTON_DUP;
	else if (b & PAD_BUTTON_DOWN)
		pad_status &= PSX_BUTTON_DDOWN;
	if (b & PAD_BUTTON_LEFT)
		pad_status &= PSX_BUTTON_DLEFT;
	else if (b & PAD_BUTTON_RIGHT)
		pad_status &= PSX_BUTTON_DRIGHT;
	/* Shoulder Buttons, L1 = L, R1 = R, L2 = L+Z, R2 = R+Z */
	if ((b & PAD_TRIGGER_Z) && (b & PAD_TRIGGER_R))
		pad_status &= PSX_BUTTON_R2;
	if ((b & PAD_TRIGGER_Z) && (b & PAD_TRIGGER_L))
		pad_status &= PSX_BUTTON_L2;
	if ((!(b & PAD_TRIGGER_Z)) && (b & PAD_TRIGGER_R))
		pad_status &= PSX_BUTTON_R1;
	if ((!(b & PAD_TRIGGER_Z)) && (b & PAD_TRIGGER_L))
		pad_status &= PSX_BUTTON_L1;

#ifdef HW_RVL
  int h = WPAD_ButtonsHeld(1);
	/* PAD Buttons */
	if (h & WPAD_CLASSIC_BUTTON_PLUS)
		pad_status &= PSX_BUTTON_START;
	if (h & WPAD_CLASSIC_BUTTON_MINUS)
		pad_status &= PSX_BUTTON_SELECT;
	if (h & WPAD_CLASSIC_BUTTON_B)
		pad_status &= PSX_BUTTON_CROSS;
	if (h & WPAD_CLASSIC_BUTTON_Y)
		pad_status &= PSX_BUTTON_SQUARE;
	if (h & WPAD_CLASSIC_BUTTON_A)
		pad_status &= PSX_BUTTON_CIRCLE;
	if (h & WPAD_CLASSIC_BUTTON_X)
		pad_status &= PSX_BUTTON_TRIANGLE;
	if (h & WPAD_CLASSIC_BUTTON_UP)
		pad_status &= PSX_BUTTON_DUP;
	else if (h & WPAD_CLASSIC_BUTTON_DOWN)
		pad_status &= PSX_BUTTON_DDOWN;
	if (h & WPAD_CLASSIC_BUTTON_LEFT)
		pad_status &= PSX_BUTTON_DLEFT;
	else if (h & WPAD_CLASSIC_BUTTON_RIGHT)
		pad_status &= PSX_BUTTON_DRIGHT;
	/* Shoulder Buttons */
	if (h & WPAD_CLASSIC_BUTTON_FULL_R)
		pad_status &= PSX_BUTTON_R2;
	if (h & WPAD_CLASSIC_BUTTON_FULL_L)
		pad_status &= PSX_BUTTON_L2;
	if (h & WPAD_CLASSIC_BUTTON_ZR)
		pad_status &= PSX_BUTTON_R1;
	if (h & WPAD_CLASSIC_BUTTON_ZL)
		pad_status &= PSX_BUTTON_L1;
#endif

	if(!controllerType) {
    // Standard Pad
    pad->controllerType = PSE_PAD_TYPE_STANDARD;
    //allow GC analog pad to be used for digital
    if(PAD_StickX(1) < -50)
      pad_status &= PSX_BUTTON_DLEFT;
    else if(PAD_StickX(1) > 50)
      pad_status &= PSX_BUTTON_DRIGHT;
    if(PAD_StickY(1) < -50)
      pad_status &= PSX_BUTTON_DDOWN;
    else if(PAD_StickY(1) > 50)
      pad_status &= PSX_BUTTON_DUP;
#ifdef HW_RVL
    //allow Classic Controller analog pad to be used for digital
    if(WPAD_Stick(1,0,0) < -50)
      pad_status &= PSX_BUTTON_DLEFT;
    else if(WPAD_Stick(1,0,0) > 50)
      pad_status &= PSX_BUTTON_DRIGHT;
    if(WPAD_Stick(1,0,1) < -50)
      pad_status &= PSX_BUTTON_DDOWN;
    else if(WPAD_Stick(1,0,1) > 50)
      pad_status &= PSX_BUTTON_DUP;
#endif
  }

  else if(controllerType==1) {
  	//adjust values by 128 cause psx values in range 0-255 where 128 is center position
#ifndef HW_RVL
  	pad->leftJoyX  = (u8)(PAD_StickX(1)+127) & 0xFF;				//analog stick
  	pad->leftJoyY  = (u8)(-PAD_StickY(1)+127) & 0xFF;
  	pad->rightJoyX = (u8)(PAD_SubStickX(1)+127) & 0xFF;			//C-stick (Left JoyStick)
  	pad->rightJoyY = (u8)(PAD_SubStickY(1)+127) & 0xFF;
#else
  	pad->leftJoyX  = ((u8)(PAD_StickX(1)+127) + (u8)(WPAD_Stick(1,0,0)+127)) & 0xFF;
  	pad->leftJoyY  = ((u8)(-PAD_StickY(1)+127) + (u8)(-WPAD_Stick(1,0,1)+127)) & 0xFF;
  	pad->rightJoyX = ((u8)(PAD_SubStickX(1)+127) + (u8)(WPAD_Stick(1,1,0)+127)) & 0xFF;
  	pad->rightJoyY = ((u8)(PAD_SubStickY(1)+127) + (u8)(WPAD_Stick(1,1,1)+127)) & 0xFF;
#endif
  	pad->controllerType = PSE_PAD_TYPE_ANALOGPAD; 	// Analog Pad  (Right JoyStick)
	}

	pad->buttonStatus = pad_status;					//Copy Buttons
	return PSE_PAD_ERR_SUCCESS;
}
