#include <stdio.h>
#include <math.h>
#include <gccore.h>
#include <ogcsys.h>
#include <stdlib.h>
#include <fat.h>
#include <unistd.h>
#include <sys/dir.h>

#include <string.h>
#ifdef USE_GUI
#include "../gui/GUI.h"
#define MAXLINES 16
#define PRINT GUI_print
#else
#define MAXLINES 30
#define PRINT printf
#endif

#ifdef USE_GUI
#define CLEAR() GUI_clear()
#else
#define CLEAR() printf("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n")
#endif

#ifdef HW_RVL
#include <wiiuse/wpad.h>
#endif
 
typedef struct {
	char name[MAXPATHLEN];
	int  size;
	int  attr;
} dir_ent;

static dir_ent* dir;

char* textFileBrowser(char* directory){
	// Set everything up to read
	DIR_ITER* dp = diropen(directory);
	if(!dp){ return NULL; }
	struct stat fstat;
	char filename[MAXPATHLEN];
	char buffer[96];
	int num_entries = 0;

	// Read each entry of the directory
	while( dirnext(dp, filename, &fstat) == 0 ){
		// Make sure we have room for this one
		dir = realloc( dir, (num_entries + 1) * sizeof(dir_ent) ); 
		strcpy(dir[num_entries].name, filename);
		dir[num_entries].size   = fstat.st_size;
		dir[num_entries].attr   = fstat.st_mode;
		++num_entries;
	}
	
	dirclose(dp);
	
	int currentSelection = (num_entries > 2) ? 2 : 1;
	int p = 0;
while(1){
		CLEAR();
		sprintf(buffer, "browsing %s:\n\n", directory);
		PRINT(buffer);
		int i = MIN(MAX(0,currentSelection-7),MAX(0,num_entries-14));
		int max = MIN(num_entries, MAX(currentSelection+7,14));
		for(; i<max; ++i){
			if(i == currentSelection)
				sprintf(buffer, "*");
			else	sprintf(buffer, " ");
				sprintf(buffer, "%s\t%-32s\t%s\n", buffer,
						dir[i].name, (dir[i].attr&S_IFDIR) ? "DIR" : "");
			PRINT(buffer);
		}
		
		/*** Wait for A/up/down press ***/
		p = PAD_ButtonsHeld(0);
		while (!(p & PAD_BUTTON_A) && !(p & PAD_BUTTON_UP) && !(p & PAD_BUTTON_DOWN))
		{
			p = PAD_ButtonsHeld(0);
#ifdef HW_RVL
			if(WPAD_ButtonsHeld(0) & WPAD_CLASSIC_BUTTON_UP) p |= PAD_BUTTON_UP;
			else if(WPAD_ButtonsHeld(0) & WPAD_CLASSIC_BUTTON_DOWN) p |= PAD_BUTTON_DOWN;
			if(WPAD_ButtonsHeld(0) & WPAD_CLASSIC_BUTTON_A) p |= PAD_BUTTON_A;
#endif
		}

		if(p & PAD_BUTTON_UP)   currentSelection = (--currentSelection < 0) ? num_entries-1 : currentSelection;
		else if(p & PAD_BUTTON_DOWN) currentSelection = (currentSelection + 1) % num_entries;
		if(p & PAD_BUTTON_A){
			if(dir[currentSelection].attr & S_IFDIR){
				char newDir[MAXPATHLEN];
				sprintf(newDir, "%s/%s", directory, dir[currentSelection].name);
				free(dir);
				dir = NULL;
				CLEAR();
				sprintf(buffer,"MOVING TO %s.\nPress B to continue.\n",newDir);
				PRINT(buffer);
#ifdef HW_RVL
				while (!(PAD_ButtonsHeld(0) & PAD_BUTTON_B) && !(WPAD_ButtonsHeld(0) & WPAD_CLASSIC_BUTTON_B));
#else
				while (!(PAD_ButtonsHeld(0) & PAD_BUTTON_B));
#endif
				return textFileBrowser(newDir);
			} else {
				char *newDir = malloc(MAXPATHLEN);
				sprintf(newDir, "%s/%s", directory, dir[currentSelection].name);
				free(dir);
				dir = NULL;
				CLEAR();
				sprintf(buffer,"SELECTING %s.\nPress B to continue.\n",newDir);
				PRINT(buffer);
#ifdef HW_RVL
				while (!(PAD_ButtonsHeld(0) & PAD_BUTTON_B) && !(WPAD_ButtonsHeld(0) & WPAD_CLASSIC_BUTTON_B));
#else
				while (!(PAD_ButtonsHeld(0) & PAD_BUTTON_B));
#endif
				return newDir;
			}
		}
		/*** Wait for button release ***/
		while (PAD_ButtonsHeld(0));
#ifdef HW_RVL
		while (WPAD_ButtonsHeld(0));
#endif
	}
}


